@@ -1,17 +1,24 @@
 # Copyright Contributors to the Pyro project.
 # SPDX-License-Identifier: Apache-2.0
 
-import inspect
+from functools import lru_cache
+import warnings
 
+from multipledispatch import dispatch
 import numpy as np
 
+import jax
 import jax.numpy as jnp
 from tensorflow_probability.substrates.jax import bijectors as tfb, distributions as tfd
 
 import numpyro.distributions as numpyro_dist
-from numpyro.distributions import Distribution as NumPyroDistribution, constraints
+from numpyro.distributions import (
+    Distribution as NumPyroDistribution,
+    constraints,
+    kl_divergence,
+)
 from numpyro.distributions.transforms import Transform, biject_to
-from numpyro.util import not_jax_tracer
+from numpyro.util import find_stack_level, not_jax_tracer
 
 
 def _get_codomain(bijector):
@@ -105,18 +112,83 @@ def _transform_to_bijector_constraint(constraint):
     return BijectorTransform(constraint.bijector)
 
 
+def _onehot_enumerate_support(self, expand=True):
+    n = self.event_shape[-1]
+    values = jnp.identity(n, dtype=jnp.result_type(self.dtype))
+    values = values.reshape((n,) + (1,) * len(self.batch_shape) + (n,))
+    if expand:
+        values = jnp.broadcast_to(values, (n,) + self.batch_shape + (n,))
+    return values
+
+
 class _TFPDistributionMeta(type(NumPyroDistribution)):
+    @lru_cache(maxsize=None)
     def __getitem__(cls, tfd_class):
         assert issubclass(tfd_class, tfd.Distribution)
 
+        tfd_class_name = tfd_class.__name__
+
         def init(self, *args, **kwargs):
+            warnings.warn(
+                "Importing distributions from numpyro.contrib.tfp.distributions is "
+                "deprecated. You should import distributions directly from "
+                "tensorflow_probability.substrates.jax.distributions instead.",
+                FutureWarning,
+                stacklevel=find_stack_level(),
+            )
             self.tfp_dist = tfd_class(*args, **kwargs)
 
-        init.__signature__ = inspect.signature(tfd_class.__init__)
-
-        _PyroDist = type(tfd_class.__name__, (TFPDistribution,), {})
-        _PyroDist.tfd_class = tfd_class
+        _PyroDist = type(tfd_class_name, (TFPDistribution,), {})
         _PyroDist.__init__ = init
+
+        if tfd_class is tfd.InverseGamma:
+            _PyroDist.arg_constraints = {
+                "concentration": constraints.positive,
+                "scale": constraints.positive,
+            }
+        elif tfd_class is tfd.OneHotCategorical:
+            _PyroDist.arg_constraints = {"logits": constraints.real_vector}
+            _PyroDist.has_enumerate_support = True
+            _PyroDist.support = constraints.simplex
+            _PyroDist.is_discrete = True
+            _PyroDist.enumerate_support = _onehot_enumerate_support
+        elif tfd_class is tfd.OrderedLogistic:
+            _PyroDist.arg_constraints = {
+                "cutpoints": constraints.ordered_vector,
+                "loc": constraints.real,
+            }
+        elif tfd_class is tfd.Pareto:
+            _PyroDist.arg_constraints = {
+                "concentration": constraints.positive,
+                "scale": constraints.positive,
+            }
+        elif tfd_class is tfd.TruncatedNormal:
+            _PyroDist.arg_constraints = {
+                "low": constraints.real,
+                "high": constraints.real,
+                "loc": constraints.real,
+                "scale": constraints.positive,
+            }
+        elif tfd_class is tfd.TruncatedCauchy:
+            _PyroDist.arg_constraints = {
+                "low": constraints.real,
+                "high": constraints.real,
+                "loc": constraints.real,
+                "scale": constraints.positive,
+            }
+        else:
+            if hasattr(numpyro_dist, tfd_class_name):
+                numpyro_dist_class = getattr(numpyro_dist, tfd_class_name)
+                # resolve FooProbs/FooLogits namespaces
+                numpyro_dist_class = getattr(
+                    numpyro_dist, f"{tfd_class_name}Logits", numpyro_dist_class
+                )
+                _PyroDist.arg_constraints = numpyro_dist_class.arg_constraints
+                _PyroDist.has_enumerate_support = (
+                    numpyro_dist_class.has_enumerate_support
+                )
+                _PyroDist.enumerate_support = numpyro_dist_class.enumerate_support
+
         return _PyroDist
 
 
@@ -130,9 +202,14 @@ class TFPDistribution(NumPyroDistribution, metaclass=_TFPDistributionMeta):
 
         d = TFPDistribution[tfd.Normal](0, 1)
 
-    """
+    Note that typical use cases do not require explicitly invoking this wrapper, since
+    NumPyro wraps TFP distributions automatically under the hood in model code, e.g.::
 
-    tfd_class = None
+        from tensorflow_probability.substrates.jax import distributions as tfd
+
+        def model():
+            numpyro.sample("x", tfd.Normal(0, 1))
+    """
 
     def __getattr__(self, name):
         # return parameters from the constructor
@@ -144,11 +221,13 @@ class TFPDistribution(NumPyroDistribution, metaclass=_TFPDistributionMeta):
 
     @property
     def batch_shape(self):
-        return self.tfp_dist.batch_shape
+        # TFP shapes are special tuples that can not be used directly
+        # with lax.broadcast_shapes. So we convert them to tuple.
+        return tuple(self.tfp_dist.batch_shape)
 
     @property
     def event_shape(self):
-        return self.tfp_dist.event_shape
+        return tuple(self.tfp_dist.event_shape)
 
     @property
     def has_rsample(self):
@@ -187,41 +266,20 @@ class TFPDistribution(NumPyroDistribution, metaclass=_TFPDistributionMeta):
         # XXX: this should cover most cases
         return self.support is None
 
+    def tree_flatten(self):
+        return jax.tree_util.tree_flatten(self.tfp_dist)
 
-InverseGamma = TFPDistribution[tfd.InverseGamma]
-InverseGamma.arg_constraints = {
-    "concentration": constraints.positive,
-    "scale": constraints.positive,
-}
+    @classmethod
+    def tree_unflatten(cls, aux_data, params):
+        fn = jax.tree_util.tree_unflatten(aux_data, params)
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore", category=FutureWarning)
+            return TFPDistribution[fn.__class__](**fn.parameters)
 
 
-def _onehot_enumerate_support(self, expand=True):
-    n = self.event_shape[-1]
-    values = jnp.identity(n, dtype=jnp.result_type(self.dtype))
-    values = values.reshape((n,) + (1,) * len(self.batch_shape) + (n,))
-    if expand:
-        values = jnp.broadcast_to(values, (n,) + self.batch_shape + (n,))
-    return values
-
-
-OneHotCategorical = TFPDistribution[tfd.OneHotCategorical]
-OneHotCategorical.arg_constraints = {"logits": constraints.real_vector}
-OneHotCategorical.has_enumerate_support = True
-OneHotCategorical.support = constraints.simplex
-OneHotCategorical.is_discrete = True
-OneHotCategorical.enumerate_support = _onehot_enumerate_support
-
-OrderedLogistic = TFPDistribution[tfd.OrderedLogistic]
-OrderedLogistic.arg_constraints = {
-    "cutpoints": constraints.ordered_vector,
-    "loc": constraints.real,
-}
-
-Pareto = TFPDistribution[tfd.Pareto]
-Pareto.arg_constraints = {
-    "concentration": constraints.positive,
-    "scale": constraints.positive,
-}
+@dispatch(TFPDistribution, TFPDistribution)
+def kl_divergence(p, q):  # noqa: F811
+    return tfd.kl_divergence(p.tfp_dist, q.tfp_dist)
 
 
 __all__ = ["BijectorConstraint", "BijectorTransform", "TFPDistribution"]
@@ -234,22 +292,9 @@ for _name, _Dist in tfd.__dict__.items():
     if _Dist is tfd.Distribution:
         continue
 
-    try:
-        _PyroDist = locals()[_name]
-    except KeyError:
-        _PyroDist = TFPDistribution[_Dist]
-        _PyroDist.__module__ = __name__
-        if hasattr(numpyro_dist, _name):
-            numpyro_dist_class = getattr(numpyro_dist, _name)
-            # resolve FooProbs/FooLogits namespaces
-            if type(numpyro_dist_class).__name__ == "function":
-                if not hasattr(numpyro_dist, _name + "Logits"):
-                    continue
-                numpyro_dist_class = getattr(numpyro_dist, _name + "Logits")
-            _PyroDist.arg_constraints = numpyro_dist_class.arg_constraints
-            _PyroDist.has_enumerate_support = numpyro_dist_class.has_enumerate_support
-            _PyroDist.enumerate_support = numpyro_dist_class.enumerate_support
-        locals()[_name] = _PyroDist
+    _PyroDist = TFPDistribution[_Dist]
+    _PyroDist.__module__ = __name__
+    locals()[_name] = _PyroDist
 
     _PyroDist.__doc__ = """
     Wraps `{}.{} <https://www.tensorflow.org/probability/api_docs/python/tfp/substrates/jax/distributions/{}>`_
@@ -271,6 +316,6 @@ __doc__ = "\n\n".join(
     """.format(
             _name
         )
-        for _name in __all__[:_len_all] + sorted(__all__[_len_all:])
+        for _name in __all__[:_len_all]
     ]
 )
