@@ -26,20 +26,34 @@
 # POSSIBILITY OF SUCH DAMAGE.
 
 
-from jax import lax, ops
+from jax import lax
 import jax.nn as nn
 import jax.numpy as jnp
 import jax.random as random
 from jax.scipy.linalg import cho_solve, solve_triangular
-from jax.scipy.special import betainc, expit, gammaln, logit, multigammaln, ndtr, ndtri
+from jax.scipy.special import (
+    betainc,
+    betaln,
+    expit,
+    gammainc,
+    gammaln,
+    logit,
+    multigammaln,
+    ndtr,
+    ndtri,
+    xlog1py,
+    xlogy,
+)
 
 from numpyro.distributions import constraints
+from numpyro.distributions.discrete import _to_logits_bernoulli
 from numpyro.distributions.distribution import Distribution, TransformedDistribution
 from numpyro.distributions.transforms import (
     AffineTransform,
     CorrMatrixCholeskyTransform,
     ExpTransform,
     PowerTransform,
+    SigmoidTransform,
 )
 from numpyro.distributions.util import (
     cholesky_of_inverse,
@@ -52,8 +66,6 @@ from numpyro.distributions.util import (
     vec_to_tril_matrix,
 )
 
-EULER_MASCHERONI = 0.5772156649015328606065120900824024310421
-
 
 class Beta(Distribution):
     arg_constraints = {
@@ -72,10 +84,10 @@ class Beta(Distribution):
         )
         concentration1 = jnp.broadcast_to(concentration1, batch_shape)
         concentration0 = jnp.broadcast_to(concentration0, batch_shape)
+        super(Beta, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
         self._dirichlet = Dirichlet(
             jnp.stack([concentration1, concentration0], axis=-1)
         )
-        super(Beta, self).__init__(batch_shape=batch_shape, validate_args=validate_args)
 
     def sample(self, key, sample_shape=()):
         assert is_prng_key(key)
@@ -282,6 +294,9 @@ class Gamma(Distribution):
     def variance(self):
         return self.concentration / jnp.power(self.rate, 2)
 
+    def cdf(self, x):
+        return gammainc(self.concentration, self.rate * x)
+
 
 class Chi2(Gamma):
     arg_constraints = {"df": constraints.positive}
@@ -447,6 +462,9 @@ class InverseGamma(TransformedDistribution):
     def tree_flatten(self):
         return super(TransformedDistribution, self).tree_flatten()
 
+    def cdf(self, x):
+        return 1 - self.base_dist.cdf(1 / x)
+
 
 class Gumbel(Distribution):
     arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
@@ -476,7 +494,7 @@ class Gumbel(Distribution):
     @property
     def mean(self):
         return jnp.broadcast_to(
-            self.loc + self.scale * EULER_MASCHERONI, self.batch_shape
+            self.loc + self.scale * jnp.euler_gamma, self.batch_shape
         )
 
     @property
@@ -490,6 +508,66 @@ class Gumbel(Distribution):
         return self.loc - self.scale * jnp.log(-jnp.log(q))
 
 
+class Kumaraswamy(TransformedDistribution):
+    arg_constraints = {
+        "concentration1": constraints.positive,
+        "concentration0": constraints.positive,
+    }
+    reparametrized_params = ["concentration1", "concentration0"]
+    support = constraints.unit_interval
+    # XXX: This flag is used to approximate the Taylor expansion
+    # of KL(Kumaraswamy||Beta) following
+    # https://arxiv.org/abs/1605.06197 Formula (12)
+    # We follow the paper and set this to 10 but to get more precise KL,
+    # we can set this flag to 1000.
+    KL_KUMARASWAMY_BETA_TAYLOR_ORDER = 10
+
+    def __init__(self, concentration1, concentration0, validate_args=None):
+        self.concentration1, self.concentration0 = promote_shapes(
+            concentration1, concentration0
+        )
+        batch_shape = lax.broadcast_shapes(
+            jnp.shape(concentration1), jnp.shape(concentration0)
+        )
+        base_dist = Uniform(0, 1).expand(batch_shape)
+        transforms = [
+            PowerTransform(1 / concentration0),
+            AffineTransform(1, -1),
+            PowerTransform(1 / concentration1),
+        ]
+        super().__init__(base_dist, transforms, validate_args=validate_args)
+
+    def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
+        minval = jnp.finfo(jnp.result_type(float)).tiny
+        u = random.uniform(key, shape=sample_shape + self.batch_shape, minval=minval)
+        log_sample = jnp.log1p(-(u ** (1 / self.concentration0))) / self.concentration1
+        finfo = jnp.finfo(u)
+        return jnp.clip(jnp.exp(log_sample), a_min=finfo.tiny, a_max=1 - finfo.eps)
+
+    @validate_sample
+    def log_prob(self, value):
+        normalize_term = jnp.log(self.concentration0) + jnp.log(self.concentration1)
+        return (
+            xlogy(self.concentration1 - 1, value)
+            + xlog1py(self.concentration0 - 1, -(value ** self.concentration1))
+            + normalize_term
+        )
+
+    @property
+    def mean(self):
+        log_beta = betaln(1 + 1 / self.concentration1, self.concentration0)
+        return self.concentration0 * jnp.exp(log_beta)
+
+    @property
+    def variance(self):
+        log_beta = betaln(1 + 2 / self.concentration1, self.concentration0)
+        return self.concentration0 * jnp.exp(log_beta) - jnp.square(self.mean)
+
+    def tree_flatten(self):
+        return super(TransformedDistribution, self).tree_flatten()
+
+
 class Laplace(Distribution):
     arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
     support = constraints.real
@@ -535,7 +613,7 @@ class Laplace(Distribution):
 class LKJ(TransformedDistribution):
     r"""
     LKJ distribution for correlation matrices. The distribution is controlled by ``concentration``
-    parameter :math:`\eta` to make the probability of the correlation matrix :math:`M` propotional
+    parameter :math:`\eta` to make the probability of the correlation matrix :math:`M` proportional
     to :math:`\det(M)^{\eta - 1}`. Because of that, when ``concentration == 1``, we have a
     uniform distribution over correlation matrices.
 
@@ -751,14 +829,11 @@ class LKJCholesky(Distribution):
         w = jnp.expand_dims(jnp.sqrt(beta_sample), axis=-1) * u_hypershere
 
         # put w into the off-diagonal triangular part
-        cholesky = ops.index_add(
-            jnp.zeros(size + self.batch_shape + self.event_shape),
-            ops.index[..., 1:, :-1],
-            w,
-        )
+        cholesky = jnp.zeros(size + self.batch_shape + self.event_shape)
+        cholesky = cholesky.at[..., 1:, :-1].set(w)
         # correct the diagonal
-        # NB: we clip due to numerical precision
-        diag = jnp.sqrt(jnp.clip(1 - jnp.sum(cholesky ** 2, axis=-1), a_min=0.0))
+        # NB: beta_sample = sum(w ** 2) because norm 2 of u is 1.
+        diag = jnp.ones(cholesky.shape[:-1]).at[..., 1:].set(jnp.sqrt(1 - beta_sample))
         cholesky = cholesky + jnp.expand_dims(diag, axis=-1) * jnp.identity(
             self.dimension
         )
@@ -801,7 +876,7 @@ class LKJCholesky(Distribution):
         order = 2 * jnp.expand_dims(self.concentration, axis=-1) - order_offset
 
         # Compute unnormalized log_prob:
-        value_diag = value[..., one_to_D, one_to_D]
+        value_diag = jnp.asarray(value)[..., one_to_D, one_to_D]
         unnormalized = jnp.sum(order * jnp.log(value_diag), axis=-1)
 
         # Compute normalization constant (on the first proof of page 1999 of [1])
@@ -848,6 +923,9 @@ class LogNormal(TransformedDistribution):
     def tree_flatten(self):
         return super(TransformedDistribution, self).tree_flatten()
 
+    def cdf(self, x):
+        return self.base_dist.cdf(jnp.log(x))
+
 
 class Logistic(Distribution):
     arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
@@ -1047,6 +1125,102 @@ class MultivariateNormal(Distribution):
         return batch_shape, event_shape
 
 
+class MultivariateStudentT(Distribution):
+    arg_constraints = {
+        "df": constraints.positive,
+        "loc": constraints.real_vector,
+        "scale_tril": constraints.lower_cholesky,
+    }
+    support = constraints.real_vector
+    reparametrized_params = ["df", "loc", "scale_tril"]
+
+    def __init__(
+        self,
+        df,
+        loc=0.0,
+        scale_tril=None,
+        validate_args=None,
+    ):
+        if jnp.ndim(loc) == 0:
+            (loc,) = promote_shapes(loc, shape=(1,))
+        batch_shape = lax.broadcast_shapes(
+            jnp.shape(df), jnp.shape(loc)[:-1], jnp.shape(scale_tril)[:-2]
+        )
+        (self.df,) = promote_shapes(df, shape=batch_shape)
+        (self.loc,) = promote_shapes(loc, shape=batch_shape + loc.shape[-1:])
+        (self.scale_tril,) = promote_shapes(
+            scale_tril, shape=batch_shape + scale_tril.shape[-2:]
+        )
+        event_shape = jnp.shape(self.scale_tril)[-1:]
+        self._chi2 = Chi2(self.df)
+        super(MultivariateStudentT, self).__init__(
+            batch_shape=batch_shape,
+            event_shape=event_shape,
+            validate_args=validate_args,
+        )
+
+    def sample(self, key, sample_shape=()):
+        assert is_prng_key(key)
+        key_normal, key_chi2 = random.split(key)
+        std_normal = random.normal(
+            key_normal,
+            shape=sample_shape + self.batch_shape + self.event_shape,
+        )
+        z = self._chi2.sample(key_chi2, sample_shape)
+        y = std_normal * jnp.expand_dims(jnp.sqrt(self.df / z), -1)
+        return self.loc + jnp.squeeze(
+            jnp.matmul(self.scale_tril, y[..., jnp.newaxis]), axis=-1
+        )
+
+    @validate_sample
+    def log_prob(self, value):
+        n = self.scale_tril.shape[-1]
+        Z = (
+            jnp.log(jnp.diagonal(self.scale_tril, axis1=-2, axis2=-1)).sum(-1)
+            + 0.5 * n * jnp.log(self.df)
+            + 0.5 * n * jnp.log(jnp.pi)
+            + gammaln(0.5 * self.df)
+            - gammaln(0.5 * (self.df + n))
+        )
+        M = _batch_mahalanobis(self.scale_tril, value - self.loc)
+        return -0.5 * (self.df + n) * jnp.log1p(M / self.df) - Z
+
+    @lazy_property
+    def covariance_matrix(self):
+        # NB: this is not covariance of this distribution;
+        # the actual covariance is df / (df - 2) * covariance_matrix
+        return jnp.matmul(self.scale_tril, jnp.swapaxes(self.scale_tril, -1, -2))
+
+    @lazy_property
+    def precision_matrix(self):
+        identity = jnp.broadcast_to(
+            jnp.eye(self.scale_tril.shape[-1]), self.scale_tril.shape
+        )
+        return cho_solve((self.scale_tril, True), identity)
+
+    @property
+    def mean(self):
+        # for df <= 1. should be jnp.nan (keeping jnp.inf for consistency with scipy)
+        return jnp.broadcast_to(
+            jnp.where(jnp.expand_dims(self.df, -1) <= 1, jnp.inf, self.loc),
+            self.shape(),
+        )
+
+    @property
+    def variance(self):
+        df = jnp.expand_dims(self.df, -1)
+        var = jnp.power(self.scale_tril, 2).sum(-1) * (df / (df - 2))
+        var = jnp.where(df > 2, var, jnp.inf)
+        var = jnp.where(df <= 1, jnp.nan, var)
+        return jnp.broadcast_to(var, self.batch_shape + self.event_shape)
+
+    @staticmethod
+    def infer_shapes(df, loc, scale_tril):
+        event_shape = (scale_tril[-1],)
+        batch_shape = lax.broadcast_shapes(df, loc[:-1], scale_tril[:-2])
+        return batch_shape, event_shape
+
+
 def _batch_mv(bmat, bvec):
     r"""
     Performs a batched matrix-vector product, with compatible but different batch shapes.
@@ -1315,6 +1489,28 @@ class Pareto(TransformedDistribution):
         return super(TransformedDistribution, self).tree_flatten()
 
 
+class RelaxedBernoulliLogits(TransformedDistribution):
+    arg_constraints = {"temperature": constraints.positive, "logits": constraints.real}
+    support = constraints.unit_interval
+
+    def __init__(self, temperature, logits, validate_args=None):
+        self.temperature, self.logits = promote_shapes(temperature, logits)
+        base_dist = Logistic(logits / temperature, 1 / temperature)
+        transforms = [SigmoidTransform()]
+        super().__init__(base_dist, transforms, validate_args=validate_args)
+
+    def tree_flatten(self):
+        return super(TransformedDistribution, self).tree_flatten()
+
+
+def RelaxedBernoulli(temperature, probs=None, logits=None, validate_args=None):
+    if probs is None and logits is None:
+        raise ValueError("One of `probs` or `logits` must be specified.")
+    if probs is not None:
+        logits = _to_logits_bernoulli(probs)
+    return RelaxedBernoulliLogits(temperature, logits, validate_args=validate_args)
+
+
 class SoftLaplace(Distribution):
     """
     Smooth distribution with Laplace-like tail behavior.
@@ -1349,9 +1545,10 @@ class SoftLaplace(Distribution):
 
     def sample(self, key, sample_shape=()):
         assert is_prng_key(key)
-        u = random.uniform(
-            key, shape=sample_shape + self.batch_shape + self.event_shape
-        )
+        dtype = jnp.result_type(float)
+        finfo = jnp.finfo(dtype)
+        minval = finfo.tiny
+        u = random.uniform(key, shape=sample_shape + self.batch_shape, minval=minval)
         return self.icdf(u)
 
     # TODO: refactor validate_sample to only does validation check and use it here
@@ -1565,7 +1762,7 @@ class BetaProportion(Beta):
     """
 
     arg_constraints = {
-        "mean": constraints.unit_interval,
+        "mean": constraints.open_interval(0.0, 1.0),
         "concentration": constraints.positive,
     }
     reparametrized_params = ["mean", "concentration"]
