@@ -86,8 +86,14 @@ import jax.numpy as jnp
 
 import numpyro
 from numpyro.distributions.distribution import COERCIONS
-from numpyro.primitives import _PYRO_STACK, Messenger, apply_stack, plate
-from numpyro.util import not_jax_tracer
+from numpyro.primitives import (
+    _PYRO_STACK,
+    CondIndepStackFrame,
+    Messenger,
+    apply_stack,
+    plate,
+)
+from numpyro.util import find_stack_level, not_jax_tracer
 
 __all__ = [
     "block",
@@ -112,7 +118,7 @@ class trace(Messenger):
     Returns a handler that records the inputs and outputs at primitive calls
     inside `fn`.
 
-    **Example**
+    **Example:**
 
     .. doctest::
 
@@ -148,7 +154,7 @@ class trace(Messenger):
             # which has no name
             return
         assert not (
-            msg["type"] == "sample" and msg["name"] in self.trace
+            msg["type"] in ("sample", "deterministic") and msg["name"] in self.trace
         ), "all sites must have unique names but got `{}` duplicated".format(
             msg["name"]
         )
@@ -175,7 +181,7 @@ class replay(Messenger):
     :param fn: Python callable with NumPyro primitives.
     :param guide_trace: an OrderedDict containing execution metadata.
 
-    **Example**
+    **Example:**
 
     .. doctest::
 
@@ -201,6 +207,7 @@ class replay(Messenger):
             warnings.warn(
                 "`guide_trace` argument is deprecated. Please replace it by `trace`.",
                 FutureWarning,
+                stacklevel=find_stack_level(),
             )
         if guide_trace is not None:
             trace = guide_trace
@@ -600,14 +607,14 @@ class scope(Messenger):
     """
     This handler prepend a prefix followed by a divider to the name of sample sites.
 
-    **Example**
+    **Example:**
 
     .. doctest::
 
         >>> import numpyro
         >>> import numpyro.distributions as dist
         >>> from numpyro.handlers import scope, seed, trace
-        >>>
+
         >>> def model():
         ...     with scope(prefix="a"):
         ...         with scope(prefix="b", divider="."):
@@ -629,6 +636,14 @@ class scope(Messenger):
         if msg.get("name"):
             msg["name"] = f"{self.prefix}{self.divider}{msg['name']}"
 
+        if msg.get("cond_indep_stack"):
+            msg["cond_indep_stack"] = [
+                CondIndepStackFrame(
+                    f"{self.prefix}{self.divider}{i.name}", i.dim, i.size
+                )
+                for i in msg["cond_indep_stack"]
+            ]
+
 
 class seed(Messenger):
     """
@@ -674,11 +689,11 @@ class seed(Messenger):
 
     def __init__(self, fn=None, rng_seed=None):
         if isinstance(rng_seed, int) or (
-            isinstance(rng_seed, jnp.ndarray) and not jnp.shape(rng_seed)
+            isinstance(rng_seed, (np.ndarray, jnp.ndarray)) and not jnp.shape(rng_seed)
         ):
             rng_seed = random.PRNGKey(rng_seed)
         if not (
-            isinstance(rng_seed, jnp.ndarray)
+            isinstance(rng_seed, (np.ndarray, jnp.ndarray))
             and rng_seed.dtype == jnp.uint32
             and rng_seed.shape == (2,)
         ):
@@ -711,6 +726,10 @@ class substitute(Messenger):
     replaced by the value returned from the call to `substitute_fn`
     for the given site.
 
+    .. note:: This handler is mainly used for internal algorithms.
+        For conditioning a generative model on observed data, please
+        using the :class:`condition` handler.
+
     :param fn: Python callable with NumPyro primitives.
     :param dict data: dictionary of `numpy.ndarray` values keyed by
         site names.
@@ -827,6 +846,7 @@ class do(Messenger):
                     "Attempting to intervene on variable {} multiple times,"
                     "this is almost certainly incorrect behavior".format(msg["name"]),
                     RuntimeWarning,
+                    stacklevel=find_stack_level(),
                 )
             msg["_intervener_id"] = self._intervener_id
 
