@@ -6,7 +6,6 @@ from collections import OrderedDict, namedtuple
 from jax import grad, jacfwd, random, value_and_grad, vmap
 from jax.flatten_util import ravel_pytree
 import jax.numpy as jnp
-from jax.ops import index_update
 from jax.scipy.linalg import solve_triangular
 from jax.scipy.special import expit
 from jax.tree_util import tree_flatten, tree_map, tree_multimap
@@ -167,7 +166,7 @@ def welford_covariance(diagonal=True):
 
         mean = jnp.zeros(shape[-1])
         m2 = jnp.zeros(shape)
-        n = 0
+        n = jnp.array(0, dtype=jnp.result_type(int))
         return mean, m2, n
 
     def update_fn(sample, state):
@@ -541,7 +540,7 @@ def warmup_adapter(
         find_reasonable_step_size = identity
     ss_init, ss_update = dual_averaging()
     mm_init, mm_update, mm_final = welford_covariance(diagonal=not dense_mass)
-    adaptation_schedule = jnp.array(build_adaptation_schedule(num_adapt_steps))
+    adaptation_schedule = build_adaptation_schedule(num_adapt_steps)
     num_windows = len(adaptation_schedule)
 
     def init_fn(
@@ -678,7 +677,7 @@ def warmup_adapter(
                 identity,
             )
 
-        t_at_window_end = t == adaptation_schedule[window_idx, 1]
+        t_at_window_end = t == jnp.asarray(adaptation_schedule)[window_idx, 1]
         window_idx = jnp.where(t_at_window_end, window_idx + 1, window_idx)
         state = HMCAdaptState(
             step_size,
@@ -1040,10 +1039,7 @@ def _iterative_build_subtree(
         r_ckpts, r_sum_ckpts = cond(
             leaf_idx % 2 == 0,
             (r_ckpts, r_sum_ckpts),
-            lambda x: (
-                index_update(x[0], ckpt_idx_max, r),
-                index_update(x[1], ckpt_idx_max, r_sum),
-            ),
+            lambda x: (x[0].at[ckpt_idx_max].set(r), x[1].at[ckpt_idx_max].set(r_sum)),
             (r_ckpts, r_sum_ckpts),
             identity,
         )
