@@ -4,12 +4,15 @@
 from collections import namedtuple
 from contextlib import contextmanager
 from functools import partial
+from typing import Callable, Dict, List, Optional
 import warnings
 
 import numpy as np
 
-from jax import device_get, jacfwd, lax, random, value_and_grad
+import jax
+from jax import device_get, jacfwd, lax, random, tree_flatten, value_and_grad
 from jax.flatten_util import ravel_pytree
+from jax.lax import broadcast_shapes
 import jax.numpy as jnp
 from jax.tree_util import tree_map
 
@@ -19,7 +22,13 @@ from numpyro.distributions.transforms import biject_to
 from numpyro.distributions.util import is_identically_one, sum_rightmost
 from numpyro.handlers import condition, replay, seed, substitute, trace
 from numpyro.infer.initialization import init_to_uniform, init_to_value
-from numpyro.util import not_jax_tracer, soft_vmap, while_loop
+from numpyro.util import (
+    _validate_model,
+    find_stack_level,
+    not_jax_tracer,
+    soft_vmap,
+    while_loop,
+)
 
 __all__ = [
     "find_valid_initial_params",
@@ -60,6 +69,18 @@ def log_density(model, model_args, model_kwargs, params):
             if intermediates:
                 log_prob = site["fn"].log_prob(value, intermediates)
             else:
+                guide_shape = jnp.shape(value)
+                model_shape = tuple(
+                    site["fn"].shape()
+                )  # TensorShape from tfp needs casting to tuple
+                try:
+                    broadcast_shapes(guide_shape, model_shape)
+                except ValueError:
+                    raise ValueError(
+                        "Model and guide shapes disagree at site: '{}': {} vs {}".format(
+                            site["name"], model_shape, guide_shape
+                        )
+                    )
                 log_prob = site["fn"].log_prob(value)
 
             if (scale is not None) and (not is_identically_one(scale)):
@@ -313,7 +334,7 @@ def find_valid_initial_params(
                 if (
                     v["type"] == "sample"
                     and not v["is_observed"]
-                    and not v["fn"].is_discrete
+                    and not v["fn"].support.is_discrete
                 ):
                     constrained_values[k] = v["value"]
                     with helpful_support_errors(v):
@@ -388,16 +409,34 @@ def _get_model_transforms(model, model_args=(), model_kwargs=None):
     has_enumerate_support = False
     for k, v in model_trace.items():
         if v["type"] == "sample" and not v["is_observed"]:
-            if v["fn"].is_discrete:
+            if v["fn"].support.is_discrete:
+                enum_type = v["infer"].get("enumerate")
+                if enum_type is not None and (enum_type != "parallel"):
+                    raise RuntimeError(
+                        "This algorithm might only work for discrete sites with"
+                        f" enumerate marked 'parallel'. But the site {k} is marked"
+                        f" as '{enum_type}'."
+                    )
                 has_enumerate_support = True
                 if not v["fn"].has_enumerate_support:
+                    dist_name = type(v["fn"]).__name__
                     raise RuntimeError(
-                        "MCMC only supports continuous sites or discrete sites "
-                        f"with enumerate support, but got {type(v['fn']).__name__}."
+                        "This algorithm might only work for discrete sites with"
+                        f" enumerate support. But the {dist_name} distribution at"
+                        f" site {k} does not have enumerate support."
+                    )
+                if enum_type is None:
+                    warnings.warn(
+                        "Some algorithms will automatically enumerate the discrete"
+                        f" latent site {k} of your model. In the future,"
+                        " enumerated sites need to be marked with"
+                        " `infer={'enumerate': 'parallel'}`.",
+                        FutureWarning,
+                        stacklevel=find_stack_level(),
                     )
             else:
                 support = v["fn"].support
-                with helpful_support_errors(v):
+                with helpful_support_errors(v, raise_warnings=True):
                     inv_transforms[k] = biject_to(support)
                 # XXX: the following code filters out most situations with dynamic supports
                 args = ()
@@ -413,6 +452,16 @@ def _get_model_transforms(model, model_args=(), model_kwargs=None):
     return inv_transforms, replay_model, has_enumerate_support, model_trace
 
 
+def _partial_args_kwargs(fn, *args, **kwargs):
+    """Returns a partial function of `fn` and args, kwargs."""
+    return partial(fn, args, kwargs)
+
+
+def _drop_args_kwargs(fn, *args, **kwargs):
+    """Returns the input function `fn`, ignoring args and kwargs."""
+    return fn
+
+
 def get_potential_fn(
     model,
     inv_transforms,
@@ -447,20 +496,20 @@ def get_potential_fn(
         `deterministic` sites in the model.
     """
     if dynamic_args:
-
-        def potential_fn(*args, **kwargs):
-            return partial(potential_energy, model, args, kwargs, enum=enum)
-
-        def postprocess_fn(*args, **kwargs):
-            if replay_model:
-                # XXX: we seed to sample discrete sites (but not collect them)
-                model_ = seed(model.fn, 0) if enum else model
-                return partial(
-                    constrain_fn, model_, args, kwargs, return_deterministic=True
-                )
-            else:
-                return partial(transform_fn, inv_transforms)
-
+        potential_fn = partial(
+            _partial_args_kwargs, partial(potential_energy, model, enum=enum)
+        )
+        if replay_model:
+            # XXX: we seed to sample discrete sites (but not collect them)
+            model_ = seed(model.fn, 0) if enum else model
+            postprocess_fn = partial(
+                _partial_args_kwargs,
+                partial(constrain_fn, model, return_deterministic=True),
+            )
+        else:
+            postprocess_fn = partial(
+                _drop_args_kwargs, partial(transform_fn, inv_transforms)
+            )
     else:
         model_kwargs = {} if model_kwargs is None else model_kwargs
         potential_fn = partial(
@@ -499,23 +548,6 @@ def _guess_max_plate_nesting(model_trace):
     return max_plate_nesting
 
 
-# TODO: follow pyro.util.check_site_shape logics for more complete validation
-def _validate_model(model_trace):
-    # XXX: this validates plate statements under `enum`
-    sites = [site for site in model_trace.values() if site["type"] == "sample"]
-
-    for site in sites:
-        batch_dims = len(site["fn"].batch_shape)
-        if site.get("_control_flow_done", False):
-            batch_dims = batch_dims - 1  # remove time dimension under scan
-        plate_dims = -min([0] + [frame.dim for frame in site["cond_indep_stack"]])
-        assert (
-            plate_dims >= batch_dims
-        ), "Missing plate statement for batch dimensions at site {}".format(
-            site["name"]
-        )
-
-
 def initialize_model(
     rng_key,
     model,
@@ -586,7 +618,9 @@ def initialize_model(
     constrained_values = {
         k: v["value"]
         for k, v in model_trace.items()
-        if v["type"] == "sample" and not v["is_observed"] and not v["fn"].is_discrete
+        if v["type"] == "sample"
+        and not v["is_observed"]
+        and not v["fn"].support.is_discrete
     }
 
     if has_enumerate_support:
@@ -594,8 +628,10 @@ def initialize_model(
 
         if not isinstance(model, enum):
             max_plate_nesting = _guess_max_plate_nesting(model_trace)
-            _validate_model(model_trace)
+            _validate_model(model_trace, plate_warning="error")
             model = enum(config_enumerate(model), -max_plate_nesting - 1)
+    else:
+        _validate_model(model_trace, plate_warning="loose")
 
     potential_fn, postprocess_fn = get_potential_fn(
         model,
@@ -734,7 +770,7 @@ def _predictive(
     num_samples = int(np.prod(batch_shape))
     if num_samples > 1:
         rng_key = random.split(rng_key, num_samples)
-    rng_key = rng_key.reshape(batch_shape + (2,))
+    rng_key = rng_key.reshape((*batch_shape, 2))
     chunk_size = num_samples if parallel else 1
     return soft_vmap(
         single_prediction, (rng_key, posterior_samples), len(batch_shape), chunk_size
@@ -766,33 +802,41 @@ class Predictive(object):
         Note that this requires ``funsor`` installation.
     :param bool parallel: whether to predict in parallel using JAX vectorized map :func:`jax.vmap`.
         Defaults to False.
-    :param batch_ndims: the number of batch dimensions in posterior samples. Some usages:
+    :param batch_ndims: the number of batch dimensions in posterior samples or parameters. If `None` defaults
+        to 0 if guide is set (i.e. not `None`) and 1 otherwise. Usages for batched posterior samples:
 
         + set `batch_ndims=0` to get prediction for 1 single sample
 
         + set `batch_ndims=1` to get prediction for `posterior_samples`
-          with shapes `(num_samples x ...)`
+          with shapes `(num_samples x ...)` (same as`batch_ndims=None` with `guide=None`)
 
         + set `batch_ndims=2` to get prediction for `posterior_samples`
           with shapes `(num_chains x N x ...)`. Note that if `num_samples`
           argument is not None, its value should be equal to `num_chains x N`.
 
+        Usages for batched parameters:
+
+        + set `batch_ndims=0` to get 1 sample from the guide and parameters (same as `batch_ndims=None` with guide)
+
+        + set `batch_ndims=1` to get predictions from a one dimensional batch of the guide and parameters
+          with shapes `(num_samples x batch_size x ...)`
+
     :return: dict of samples from the predictive distribution.
 
     **Example:**
 
-    Given a model
+    Given a model::
 
         def model(X, y=None):
             ...
             return numpyro.sample("obs", likelihood, obs=y)
 
-    you can sample from the prior predictive:
+    you can sample from the prior predictive::
 
         predictive = Predictive(model, num_samples=1000)
         y_pred = predictive(rng_key, X)["obs"]
 
-    If you also have posterior samples, you can sample from the posterior predictive:
+    If you also have posterior samples, you can sample from the posterior predictive::
 
         predictive = Predictive(model, posterior_samples=posterior_samples)
         y_pred = predictive(rng_key, X)["obs"]
@@ -803,22 +847,26 @@ class Predictive(object):
 
     def __init__(
         self,
-        model,
-        posterior_samples=None,
+        model: Callable,
+        posterior_samples: Optional[Dict] = None,
         *,
-        guide=None,
-        params=None,
-        num_samples=None,
-        return_sites=None,
-        infer_discrete=False,
-        parallel=False,
-        batch_ndims=1,
+        guide: Optional[Callable] = None,
+        params: Optional[Dict] = None,
+        num_samples: Optional[int] = None,
+        return_sites: Optional[List[str]] = None,
+        infer_discrete: bool = False,
+        parallel: bool = False,
+        batch_ndims: Optional[int] = None,
     ):
         if posterior_samples is None and num_samples is None:
             raise ValueError(
                 "Either posterior_samples or num_samples must be specified."
             )
 
+        batch_ndims = (
+            batch_ndims if batch_ndims is not None else 1 if guide is None else 0
+        )
+
         posterior_samples = {} if posterior_samples is None else posterior_samples
 
         prototype_site = batch_shape = batch_size = None
@@ -840,6 +888,7 @@ class Predictive(object):
                             batch_size, num_samples, batch_size
                         ),
                         UserWarning,
+                        stacklevel=find_stack_level(),
                     )
                 num_samples = batch_size
 
@@ -865,21 +914,12 @@ class Predictive(object):
         self.batch_ndims = batch_ndims
         self._batch_shape = batch_shape
 
-    def __call__(self, rng_key, *args, **kwargs):
-        """
-        Returns dict of samples from the predictive distribution. By default, only sample sites not
-        contained in `posterior_samples` are returned. This can be modified by changing the
-        `return_sites` keyword argument of this :class:`Predictive` instance.
-
-        :param jax.random.PRNGKey rng_key: random key to draw samples.
-        :param args: model arguments.
-        :param kwargs: model kwargs.
-        """
+    def _call_with_params(self, rng_key, params, args, kwargs):
         posterior_samples = self.posterior_samples
         if self.guide is not None:
             rng_key, guide_rng_key = random.split(rng_key)
             # use return_sites='' as a special signal to return all sites
-            guide = substitute(self.guide, self.params)
+            guide = substitute(self.guide, params)
             posterior_samples = _predictive(
                 guide_rng_key,
                 guide,
@@ -903,6 +943,29 @@ class Predictive(object):
             model_kwargs=kwargs,
         )
 
+    def __call__(self, rng_key, *args, **kwargs):
+        """
+        Returns dict of samples from the predictive distribution. By default, only sample sites not
+        contained in `posterior_samples` are returned. This can be modified by changing the
+        `return_sites` keyword argument of this :class:`Predictive` instance.
+
+        :param jax.random.PRNGKey rng_key: random key to draw samples.
+        :param args: model arguments.
+        :param kwargs: model kwargs.
+        """
+        if self.batch_ndims == 0 or self.params == {} or self.guide is None:
+            return self._call_with_params(rng_key, self.params, args, kwargs)
+        elif self.batch_ndims == 1:  # batch over parameters
+            batch_size = jnp.shape(tree_flatten(self.params)[0][0])[0]
+            rng_keys = random.split(rng_key, batch_size)
+            return jax.vmap(
+                partial(self._call_with_params, args=args, kwargs=kwargs),
+                in_axes=0,
+                out_axes=1,
+            )(rng_keys, self.params)
+        else:
+            raise NotImplementedError
+
 
 def log_likelihood(
     model, posterior_samples, *args, parallel=False, batch_ndims=1, **kwargs
@@ -961,12 +1024,28 @@ def log_likelihood(
 
 
 @contextmanager
-def helpful_support_errors(site):
+def helpful_support_errors(site, raise_warnings=False):
+    name = site["name"]
+    support = getattr(site["fn"], "support", None)
+    if isinstance(support, constraints.independent):
+        support = support.base_constraint
+
+    # Warnings
+    if raise_warnings:
+        if support is constraints.circular:
+            msg = (
+                f"Continuous inference poorly handles circular sample site '{name}'. "
+                + "Consider using VonMises distribution together with "
+                + "a reparameterizer, e.g. "
+                + f"numpyro.handlers.reparam(config={{'{name}': CircularReparam()}})."
+            )
+            warnings.warn(msg, UserWarning, stacklevel=find_stack_level())
+
+    # Exceptions
     try:
         yield
     except NotImplementedError as e:
-        name = site["name"]
-        support_name = repr(site["fn"].support).lower()
+        support_name = repr(support).lower()
         if "integer" in support_name or "boolean" in support_name:
             # TODO: mention enumeration when it is supported in SVI
             raise ValueError(
