@@ -22,6 +22,7 @@ EXAMPLES = [
     "capture_recapture.py --num-samples 4 --num-warmup 1 -m 3",
     "capture_recapture.py --num-samples 4 --num-warmup 1 -m 5",
     "covtype.py --algo HMC --num-samples 10 --num-warmup 10",
+    "dais_demo.py --num-svi-steps 10 --num-samples 10 --num-warmup 10",
     "gaussian_shells.py --num-samples 100",
     "gaussian_shells.py --num-samples 100 --enum",
     "gp.py --num-samples 10 --num-warmup 10 --num-chains 2",
@@ -34,19 +35,31 @@ EXAMPLES = [
     "hmm_enum.py -m 3 -t 3 -d 4 --num-warmup 1 -n 4",
     "hmm_enum.py -m 4 -t 3 -d 4 --num-warmup 1 -n 4",
     "hmm_enum.py -m 6 -t 4 -d 3 --num-warmup 1 -n 4",
-    "hsgp.py --model t --num-samples 10 --num-warmup 10 --num-chains 2",
-    "hsgp.py --model ty --num-samples 10 --num-warmup 10 --num-chains 2",
-    "hsgp.py --model tyw --num-samples 10 --num-warmup 10 --num-chains 2",
-    "hsgp.py --model tywd --num-samples 10 --num-warmup 10 --num-chains 2",
+    "horseshoe_regression.py --num-samples 10 --num-warmup 10 --num-data 5",
+    "hsgp.py --num-samples 10 --num-warmup 10 --num-chains 2",
     "minipyro.py",
     "neutra.py --num-samples 100 --num-warmup 100",
     "ode.py --num-samples 100 --num-warmup 100 --num-chains 1",
-    "prodlda.py --num-steps 10 --hidden 10 --nn-framework flax",
-    "prodlda.py --num-steps 10 --hidden 10 --nn-framework haiku",
+    pytest.param(
+        "prodlda.py --num-steps 10 --hidden 10 --nn-framework flax",
+        marks=pytest.mark.skipif(
+            "CI" in os.environ, reason="This example requires a lot of memory."
+        ),
+    ),
+    pytest.param(
+        "prodlda.py --num-steps 10 --hidden 10 --nn-framework haiku",
+        marks=pytest.mark.skipif(
+            "CI" in os.environ, reason="This example requires a lot of memory."
+        ),
+    ),
     "sparse_regression.py --num-samples 10 --num-warmup 10 --num-data 10 --num-dimensions 10",
+    "ssbvm_mixture.py --num-samples 10 --num-warmup 10",
     "stochastic_volatility.py --num-samples 100 --num-warmup 100",
     "ucbadmit.py --num-chains 2",
     "vae.py -n 1",
+    "ar2.py --num-samples 10 --num-warmup 10 --num-chains 2",
+    "ar2.py --num-samples 10 --num-warmup 10 --num-chains 2 --num-data 10 --unroll-loop",
+    "holt_winters.py --T 4 --num-samples 10 --num-warmup 10 --num-chains 2",
 ]
 
 
