@@ -12,7 +12,7 @@ import numpyro
 from numpyro import handlers
 import numpyro.distributions as dist
 from numpyro.distributions import constraints
-from numpyro.infer import SVI, Trace_ELBO
+from numpyro.infer import MCMC, NUTS, SVI, Trace_ELBO
 from numpyro.infer.util import log_density
 import numpyro.optim as optim
 from numpyro.util import not_jax_tracer, optional
@@ -22,7 +22,7 @@ from numpyro.util import not_jax_tracer, optional
 @pytest.mark.parametrize("use_jit", [False, True])
 def test_mask(mask_last, use_jit):
     N = 10
-    mask = np.ones(N, dtype=np.bool)
+    mask = np.ones(N, dtype=bool)
     mask[-mask_last] = 0
 
     def model(data, mask):
@@ -591,6 +591,26 @@ def test_scope():
     assert "b/a/x" in trace
 
 
+def test_scope_frames():
+    def model(y):
+        mu = numpyro.sample("mu", dist.Normal())
+        sigma = numpyro.sample("sigma", dist.HalfNormal())
+
+        with numpyro.plate("plate1", y.shape[0]):
+            numpyro.sample("y", dist.Normal(mu, sigma), obs=y)
+
+    scope_prefix = "scope"
+    scoped_model = handlers.scope(model, prefix=scope_prefix)
+
+    obs = np.random.normal(size=(10,))
+
+    trace = handlers.trace(handlers.seed(model, 0)).get_trace(obs)
+    scoped_trace = handlers.trace(handlers.seed(scoped_model, 0)).get_trace(obs)
+
+    assert trace["y"]["cond_indep_stack"][0].name in trace
+    assert scoped_trace[f"{scope_prefix}/y"]["cond_indep_stack"][0].name in scoped_trace
+
+
 def test_lift():
     def model():
         loc1 = numpyro.param("loc1", 0.0)
@@ -757,4 +777,15 @@ def test_subsample_fn():
         )
 
         # test that values are not duplicated
-        assert len(set(subsamples[k])) == subsample_size
+        assert len(set(subsamples[k].copy())) == subsample_size
+
+
+def test_sites_have_unique_names():
+    def model():
+        alpha = numpyro.sample("alpha", dist.Normal())
+        numpyro.deterministic("alpha", alpha * 2)
+
+    mcmc = MCMC(NUTS(model), num_chains=1, num_samples=10, num_warmup=10)
+    msg = "all sites must have unique names but got `alpha` duplicated"
+    with pytest.raises(AssertionError, match=msg):
+        mcmc.run(random.PRNGKey(0))
